/* Copyright (C) 1999 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.
   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* Handler-calling-functions */

#include "mysql_priv.h"
#include <errno.h>
#include <nisam.h>
#include <myisam.h>
#include <myisampack.h>
#ifndef NO_HEAP
#include <heap.h>
#else
#define heap_open 0				/* Undefine for initarray */
#define heap_close 0
#define heap_write 0
#define heap_update 0
#define heap_delete 0
#define heap_rsame 0
#define heap_rrnd 0
#define heap_extra 0
#define heap_rnext 0
#define heap_rprev 0
#define heap_rfirst 0
#define heap_rlast 0
#define heap_rkey 0
#endif
#ifndef NO_HASH
#include <hashdb.h>
#else
#define h_open 0				/* Undefine for initarray */
#define h_close 0
#define h_write 0
#define h_update 0
#define h_delete 0
#define h_rkey 0
#define h_rnext 0
#define h_rprev 0
#define h_rfirst 0
#define h_rlast 0
#define h_rsame 0
#define h_rrnd 0
#define h_lock_database 0
#define h_extra 0
#endif
#ifndef NO_MERGE
#include <merge.h>
#else
#define mrg_open 0				/* Undefine for initarray */
#define mrg_close 0
#define mrg_write 0
#define mrg_update 0
#define mrg_delete 0
#define mrg_rkey 0
#define mrg_rnext 0
#define mrg_rprev 0
#define mrg_rfirst 0
#define mrg_rlast 0
#define mrg_rsame 0
#define mrg_rrnd 0
#define mrg_lock_database 0
#endif
#ifndef NO_MYMERGE
#include <myisammrg.h>
#else
#define myrg_extra 0
#define myrg_open 0				/* Undefine for initarray */
#define myrg_close 0
#define myrg_write 0
#define myrg_update 0
#define myrg_delete 0
#define myrg_rlast 0
#define myrg_rsame 0
#define myrg_rrnd 0
#define myrg_lock_database 0
#define myrg_extra 0
#endif


	/* static functions defined in this file */

static int NEAR_F _ha_open(TABLE *form, my_string name,int mode,
			   int test_if_locked);
static void NEAR_F set_form_timestamp(TABLE *table, byte *record);
static int ha_err(void);
static int ha_ok(void);

ulong ha_read_count, ha_write_count, ha_delete_count, ha_update_count,
      ha_read_key_count, ha_read_first_count, ha_read_next_count, ha_read_rnd_count;

	/* Index to functions in different handlers */

typedef int (*F) ();

static	int (*funcs[][9])() = {
  { (F) h_open,		(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_open,		(F) ni_open,
    (F) mrg_open,	(F) mi_open,		(F) myrg_open },
  { (F) h_close,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_close,		(F) ni_close,
    (F) mrg_close,	(F) mi_close,		(F) myrg_close },
  { (F) h_write,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_write,		(F) ni_write,
    (F) ha_err,		(F) mi_write,		(F) ha_err },
  { (F) h_update,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_update,	(F) ni_update,
    (F) mrg_update,	(F) mi_update,		(F) myrg_update },
  { (F) h_delete,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_delete,	(F) ni_delete,
    (F) mrg_delete,	(F) mi_delete,		(F) myrg_delete },
  { (F) h_rkey,		(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rkey,		(F) ni_rkey,
    (F) ha_err,		(F) mi_rkey,		(F) ha_err },
  { (F) h_rnext,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rnext,		(F) ni_rnext,
    (F) ha_err,		(F) mi_rnext,		(F) ha_err },
  { (F) h_rprev,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rprev,		(F) ni_rprev,
    (F) ha_err,		(F) mi_rprev,		(F) ha_err },
  { (F) h_rfirst,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rfirst,	(F) ni_rfirst,
    (F) ha_err,		(F) mi_rfirst,		(F) ha_err },
  { (F) h_rlast,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rlast,		(F) ni_rlast,
    (F) ha_err,		(F) mi_rlast,		(F) ha_err },
#ifdef NOT_NEEDED
  { (F) h_rsame,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rsame,		(F) ni_rsame,
    (F) mrg_rsame,	(F) mi_rsame,		(F) myrg_rsame },
#else
  { (F) ha_err,		(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) ha_err,		(F) ha_err,
    (F) 0 },
#endif
  { (F) h_rrnd,		(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_rrnd,		(F) ni_rrnd,
    (F) mrg_rrnd,	(F) mi_rrnd,		(F) myrg_rrnd },
  { (F) h_lock_database,(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) ha_ok,		(F) ni_lock_database,
    (F) mrg_lock_database, (F) mi_lock_database,(F) myrg_lock_database },
  { (F) h_extra,	(F) ha_err,		(F) ha_err,
    (F) ha_err,		(F) heap_extra,		(F) ni_extra,
    (F) mrg_extra,	(F) mi_extra,		(F) myrg_extra }
};

ulong NEAR ha_option_flag[] = {
  0,
  /* DIAB_ISAM */
  HA_READ_NEXT+HA_READ_PREV+HA_READ_ORDER+HA_RSAME_NO_INDEX,
  /* HASH */
  HA_READ_NEXT+HA_READ_PREV+HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS+
  HA_BINARY_KEYS+HA_WRONG_ASCII_ORDER,
    /* M-ISAM */
  HA_READ_NEXT+HA_READ_PREV+HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS+
  HA_READ_ORDER+HA_LASTKEY_ORDER+HA_HAVE_KEY_READ_ONLY+HA_READ_NOT_EXACT_KEY,
    /* P-ISAM */
  HA_READ_NEXT+HA_READ_PREV+HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS+
  HA_READ_ORDER+HA_LASTKEY_ORDER+HA_REC_NOT_IN_SEQ+HA_HAVE_KEY_READ_ONLY+
  HA_READ_NOT_EXACT_KEY,
    /* RMS-ISAM */
  HA_READ_NEXT+HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS+
  HA_READ_ORDER+HA_LASTKEY_ORDER+HA_REC_NOT_IN_SEQ+HA_RSAME_NO_INDEX+
  HA_WRONG_ASCII_ORDER+HA_READ_NOT_EXACT_KEY,
    /* HEAP */
  HA_READ_RND_SAME+HA_NO_INDEX+HA_BINARY_KEYS+HA_WRONG_ASCII_ORDER+
  HA_KEYPOS_TO_RNDPOS+HA_NO_BLOBS,
    /* ISAM */
  HA_READ_NEXT+HA_READ_PREV+HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS+
  HA_READ_ORDER+HA_LASTKEY_ORDER+HA_HAVE_KEY_READ_ONLY+HA_READ_NOT_EXACT_KEY+
  HA_LONGLONG_KEYS+HA_KEY_READ_WRONG_STR + HA_DUPP_POS,
    /* MERGE-ISAM */
  HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS,
    /* MY_ISAM */
  HA_READ_NEXT+HA_READ_PREV+HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS+
  HA_READ_ORDER+HA_LASTKEY_ORDER+HA_HAVE_KEY_READ_ONLY+HA_READ_NOT_EXACT_KEY+
  HA_LONGLONG_KEYS+HA_NULL_KEY + HA_DUPP_POS ,
    /* MERGE-MY_ISAM */
  HA_READ_RND_SAME+HA_KEYPOS_TO_RNDPOS,
};


uint NEAR ha_maxrecordlength[]=
{ 0,1024,32767,32767,65535,65535,65535,65535,65535,65535,65535};
uint NEAR ha_max_keys[]=
{ 0,MAX_KEY,MAX_KEY,MAX_KEY,MAX_KEY,MAX_KEY,16,N_MAXKEY,0,MI_MAX_KEY,0 };
uint NEAR ha_max_key_parts[]=
{ 0,5,5,5,5,5,16,N_MAXKEY_SEG,0,MI_MAX_KEY_SEG,0};
uint NEAR ha_max_key_length[]=
{ N_MAX_KEY_LENGTH,N_MAX_KEY_LENGTH,N_MAX_KEY_LENGTH,N_MAX_KEY_LENGTH,
  N_MAX_KEY_LENGTH,N_MAX_KEY_LENGTH,65535,N_MAX_KEY_LENGTH,
  N_MAX_KEY_LENGTH,MI_MAX_KEY_LENGTH, MI_MAX_KEY_LENGTH };

char NEAR bas_ext[][2][FN_EXTLEN] = {
  { "","" },
  { ".DAT",".ISM"},				/* ISAM extensions */
  { ".DAT",".HSH"},				/* HASH extensions */
  { ".ISD",".ISM"},				/* M-ISAM extensions */
  { ".PSD",".PSM"},				/* P-ISAM extensions */
  { ".DAT",""},					/* RMS-ISAM extensions */
  { "",""},					/* HEAP extensions */
  { ".ISD",".ISM"},
  { ".MRG",""},
  { ".MYD",".MYI"},
  { ".MRG",""},
};

char *ha_table_type[] = {
  "", "DIAB_ISAM","HASH","MISAM","PISAM","RMS_ISAM","HEAP", "ISAM",
  "MRG_ISAM","MYISAM", "MRG_MYISAM", "?", "?", "?",NullS
};

TYPELIB ha_table_typelib= {array_elements(ha_table_type)-4,"",ha_table_type+1};

	/* Open database-handler. Try O_RDONLY if can't open as O_RDWR */
	/* Don't wait for locks if not HA_OPEN_WAIT_IF_LOCKED is set */

int ha_open(TABLE *form, my_string name, int mode, int test_if_locked)
{
  int error;
  DBUG_ENTER("ha_open");
  DBUG_PRINT("enter",("db_type: %d  db_stat: %d  mode: %d  lock_test: %d",
		      form->db_type, form->db_stat, mode, test_if_locked));

  form->file=0 ;
  form->db_capabilities=ha_option_flag[form->db_type];
  if ((error=_ha_open(form,name,mode,test_if_locked)))
  {
    if ((error == EACCES || error == EROFS) && mode == O_RDWR &&
	(form->db_stat & HA_TRY_READ_ONLY))
    {
      form->db_stat|=HA_READ_ONLY;
      error=_ha_open(form,name,O_RDONLY,test_if_locked);
    }
  }
  if (error)
  {
    my_errno=error;			/* Safeguard */
    DBUG_PRINT("error",("error: %d  errno: %d",error,errno));
  }
  DBUG_RETURN (error);
} /* ha_open */


	/* sub_function to ha_open */
	/* Name will be changed !! */

static int NEAR_F _ha_open(register TABLE *form, my_string name, int mode,
			   int test_if_locked)
{
  char name_buff[FN_REFLEN];
  form->keyfile_info.ref_length=sizeof(my_off_t); /* Length of ref */
  form->locked=F_UNLCK;				 /* Not locked */
  switch (form->db_type) {
#ifndef NO_HEAP
  case DB_TYPE_HEAP:
    {
      if (form->db_stat & (HA_OPEN_KEYFILE+HA_OPEN_RNDFILE+HA_GET_INFO))
      {
	uint key,part,parts;
	HP_KEYDEF *keydef;
	HP_KEYSEG *seg;

	for (key=parts=0 ; key < form->keys ; key++)
	  parts+=form->key_info[key].key_parts;

	if (!(keydef=(HP_KEYDEF*) my_alloca(form->keys*sizeof(HP_KEYDEF)+
					    parts*sizeof(HP_KEYSEG))))
	  return my_errno;
	seg=my_reinterpret_cast(HP_KEYSEG*) (keydef+form->keys);
	for (key=0 ; key < form->keys ; key++)
	{
	  KEY *pos=form->key_info+key;

	  keydef[key].keysegs=(uint) pos->key_parts;
	  keydef[key].flag = pos->dupp_key ? 0 : HA_NOSAME;
	  keydef[key].seg=seg;

	  for (part=0 ; part < pos->key_parts ; part++)
	  {
	    uint flag=pos->key_part[part].key_type;
	    if (!f_is_packed(flag) &&
		f_packtype(flag) == (int) FIELD_TYPE_DECIMAL &&
		!(flag & FIELDFLAG_BINARY))
	      seg->type= (int) HA_KEYTYPE_TEXT;
	    else
	      seg->type= (int) HA_KEYTYPE_BINARY;
	    seg->start=(uint) pos->key_part[part].offset;
	    seg->length=(uint) pos->key_part[part].length;
	    seg++;
	  }
	}
	form->file=(byte*) heap_open(fn_format(name_buff,name,"","",1+2),mode,
				     form->keys,keydef,
				     form->reclength,form->max_rows,
				     form->min_rows);
	my_afree((gptr) keydef);
	ha_info(form,2);
	return (!form->file ? errno : 0);
      }
    }
    break;
#endif
  case DB_TYPE_MRG_ISAM:
  case DB_TYPE_MRG_MYISAM:
    if (form->db_stat == HA_GET_INFO)
    {						/* Skipp if not open */
      form->db_stat=0;				/* because .MRG may be wrong */
      break;
    }
    /* fall through */
  case DB_TYPE_MYISAM:
  case DB_TYPE_HASH:
  case DB_TYPE_ISAM:
    {
      if (form->db_stat & (HA_OPEN_KEYFILE+HA_OPEN_RNDFILE+HA_GET_INFO))
      {
	if (!(form->file= (* ((byte *(*)(char *,int,int))
			      funcs[0][(int) form->db_type-2]))
	      (fn_format(name_buff,name,"","",2 | 4),mode,test_if_locked)))
	  return(my_errno ? my_errno : -1);
	if (!(test_if_locked == HA_OPEN_WAIT_IF_LOCKED ||
	      test_if_locked == HA_OPEN_ABORT_IF_LOCKED))
	  VOID(ha_extra(form,HA_EXTRA_NO_WAIT_LOCK));
	ha_info(form,2);
	if (!(test_if_locked & HA_OPEN_WAIT_IF_LOCKED))
	  VOID(ha_extra(form,HA_EXTRA_WAIT_LOCK));
	if (form->db_create_options &
	    (HA_OPTION_PACK_RECORD | HA_OPTION_COMPRESS_RECORD) ||
	    form->reclength < 9)
	  form->db_capabilities|=HA_REC_NOT_IN_SEQ;
	if (form->db_create_options & HA_OPTION_READ_ONLY_DATA)
	  form->db_stat|=HA_READ_ONLY;
	if (form->db_type != DB_TYPE_MISAM && form->db_type != DB_TYPE_HASH)
	  form->db_record_offset=form->reclength;

	if (form->db_type == DB_TYPE_MRG_ISAM ||
	    form->db_type == DB_TYPE_MRG_MYISAM)
	{
	  if (form->reclength != form->keyfile_info.mean_rec_length)
	  {
	    DBUG_PRINT("error",("reclength: %d  mean_rec_length: %d",
				form->reclength,
				form->keyfile_info.mean_rec_length));
	    mrg_close((MRG_INFO*) form->file);
	    return ER_WRONG_MRG_TABLE;
	  }
	}
	if (form->db_type == DB_TYPE_MYISAM ||
	    form->db_type == DB_TYPE_MRG_MYISAM)
	{
	  form->db_low_byte_first=1;
	}
      }
    }
    break;
  default:
    break;
  }
  return 0;
}


	/* close database */

int ha_close(register TABLE *form)
{
  int error=0;
  DBUG_ENTER("ha_close");

  if (form->db_stat & (HA_OPEN_KEYFILE+HA_OPEN_RNDFILE+HA_GET_INFO))
    error= (* ((int (*)(byte*)) (*funcs[1][(int) form->db_type-2])))(form->file);
  if (error)
    my_error(ER_ERROR_ON_CLOSE,MYF(ME_BELL+ME_WAITTANG),
	     form->table_name,my_errno);
  DBUG_RETURN(error);
} /* ha_close */


	/* write to database. Fix timestamp and next_number */

int ha_write(register TABLE *form, byte * buf)
{
  register int error=0;
  DBUG_ENTER("ha_write");
  pthread_mutex_lock(&LOCK_status);
  ha_write_count++;
  pthread_mutex_unlock(&LOCK_status);

  if (form->time_stamp)
    set_form_timestamp(form,buf+form->time_stamp-1);
  if (form->next_number_field && buf == form->record[0])
    update_next_number(form);
  if ((* ((int (*)(char *,byte *))(*funcs[2][(uint) form->db_type-2])))
      ((char*) form->file,buf))
    error=my_errno ? my_errno : -1;
#ifndef DBUG_OFF
  if (error && error != HA_ERR_FOUND_DUPP_KEY &&
      error != HA_ERR_FOUND_DUPP_UNIQUE)
    DBUG_PRINT("error",("errno: %d",error));
#endif
  DBUG_RETURN (error);
} /* ha_write */


	/* update database */

int ha_update(register TABLE *form, byte * old_data, byte * new_data)
{
  register int error=0;
  DBUG_ENTER("ha_update");
  pthread_mutex_lock(&LOCK_status);
  ha_update_count++;
  pthread_mutex_unlock(&LOCK_status);

  if (form->time_stamp)
    set_form_timestamp(form,new_data+form->time_stamp-1);
  if ((* ((int (*)(char *,byte*,byte*))(*funcs[3][(uint) form->db_type-2])))
      ((char*) form->file,old_data,new_data))
    error=my_errno ? my_errno : -1;
#ifndef DBUG_OFF
  if (error)
    DBUG_PRINT("error",("errno: %d",error));
#endif
  DBUG_RETURN (error);
} /* ha_update */


	/* delete from database */

int ha_delete(register TABLE *form, byte * buf)
{
  register int error=0;
  DBUG_ENTER("ha_delete");
  pthread_mutex_lock(&LOCK_status);
  ha_delete_count++;
  pthread_mutex_unlock(&LOCK_status);

  if ((* ((int (*)(char *,byte*))(*funcs[4][(uint) form->db_type-2])))
      ((char*) form->file,buf))
    error=(my_errno ? my_errno : -1);
#ifndef DBUG_OFF
  if (error)
    DBUG_PRINT("error",("errno: %d",error));
#endif
  DBUG_RETURN (error);
} /* ha_delete */


	/* read key in database */

int ha_rkey(register TABLE *form, byte * buf, int inx, byte * key,
	    uint key_len, enum ha_rkey_function find_flag)
{
  reg2 int error=0;
  DBUG_ENTER("ha_rkey");
  DBUG_PRINT("enter",("form: %lx  '%s'",form,form->table_name));
  DBUG_PRINT("enter",("index: %d  key: '%.*s'  key_len: %d  flag: %d",
		      inx,
		      (key_len ? key_len : form->key_info[inx].key_length),key,
		      key_len,find_flag));
  DBUG_DUMP("key",(char*) key,
	    (key_len ? key_len : form->key_info[inx].key_length));

  pthread_mutex_lock(&LOCK_status);
  ha_read_key_count++;
  pthread_mutex_unlock(&LOCK_status);

  switch (form->db_type) {
#if !defined(NO_HASH) || !defined(NO_HEAP)
  case DB_TYPE_HASH:
  case DB_TYPE_HEAP:
    if ((* ((int (*)(char *,byte *,int,byte*))(*funcs[5][(uint) form->db_type-2])))((char*) form->file,buf,inx,key))
      error=my_errno ? my_errno : -1;
    break;
#endif
  default:					/* M_ISAM,PISAM,RMS,ISAM */
    if ((* ((int (*)(char *,byte *,int,byte *,uint,int))(*funcs[5][(uint) form->db_type-2])))((char*) form->file,buf,inx,key,key_len,(int) find_flag))
      error=my_errno ? my_errno : -1;
    break;
  }
  if (error)
  {
    DBUG_PRINT((error == HA_ERR_KEY_NOT_FOUND ? "extra_info" : "error"),
	       ("my_errno: %d",error));
    form->status=STATUS_NOT_FOUND;
  }
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rkey */


	/* read next record from database */

int ha_rnext(register TABLE *form, byte * buf, register int inx)
{
  reg3 int error=0;
  DBUG_ENTER("ha_rnext");
  DBUG_PRINT("enter",("form: %lx  '%s'",form,form->table_name));

  pthread_mutex_lock(&LOCK_status);
  ha_read_next_count++;
  pthread_mutex_unlock(&LOCK_status);

  if ((* ((int (*)(char *,byte *,int))(*funcs[6][(uint) form->db_type-2])))((char*) form->file,buf,inx))
    error=(my_errno ? my_errno : HA_ERR_END_OF_FILE);
  if (error)
  {
    DBUG_PRINT((error == HA_ERR_END_OF_FILE ? "extra_info" : "error"),
	       ("my_errno: %d",error));
    form->status=STATUS_NOT_FOUND;
  }
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rnext */


	/* read previous record from database */

int ha_rprev(register TABLE *form, byte * buf, register int inx)
{
  reg3 int error=0;
  DBUG_ENTER("ha_rprev");

  pthread_mutex_lock(&LOCK_status);
  ha_read_next_count++;
  pthread_mutex_unlock(&LOCK_status);

  if ((* ((int (*)(char *,byte *,int))(*funcs[7][(uint) form->db_type-2])))((char*) form->file,buf,inx))
    error=my_errno ? my_errno : HA_ERR_END_OF_FILE;
  if (error)
  {
    DBUG_PRINT("error",("my_errno: %d",error));
    form->status=STATUS_NOT_FOUND;
  }
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rprev */

	/* read first record from database */

int ha_rfirst(register TABLE *form, byte * buf, int inx)
{
  register int error=0;
  DBUG_ENTER("ha_rfirst");

  pthread_mutex_lock(&LOCK_status);
  ha_read_key_count++;
  pthread_mutex_unlock(&LOCK_status);

  if ((* ((int (*)(char *,byte *,int))(*funcs[8][(uint) form->db_type-2])))((char*) form->file,buf,inx))
    error=my_errno ? my_errno : HA_ERR_END_OF_FILE;
  if (error)
  {
    DBUG_PRINT("error",("my_errno: %d",error));
    form->status=STATUS_NOT_FOUND;
  }
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rfirst */


	/* read last record from database */

int ha_rlast(register TABLE *form, byte * buf, int inx)
{
  register int error=0;
  DBUG_ENTER("ha_rlast");

  pthread_mutex_lock(&LOCK_status);
  ha_read_next_count++;
  pthread_mutex_unlock(&LOCK_status);

  if ((* ((int (*)(char *,byte *,int))(*funcs[9][(uint) form->db_type-2])))((char*)form->file,buf,inx))
    error=my_errno ? my_errno : HA_ERR_END_OF_FILE;
  if (error)
  {
    DBUG_PRINT("error",("my_errno: %d",error));
    form->status=STATUS_NOT_FOUND;
  }
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rlast */


	/* read the same record that we read last time */

#ifdef NOT_NEEDED_YET
int ha_rsame(register TABLE *form, byte * buf, int inx)
{
  register int error=0;
  DBUG_ENTER("ha_rsame");

  pthread_mutex_lock(&LOCK_status);
  ha_read_key_count++;
  pthread_mutex_unlock(&LOCK_status);

  if ((* ((int (*)(char *,byte *,int))(*funcs[10][(uint) form->db_type-2])))((char*) form->file,buf,inx))
    error=my_errno ? my_errno : -1;
  if (error)
  {
    DBUG_PRINT("error",("my_errno: %d",error));
    form->status=STATUS_NOT_FOUND;
  }
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rsame */
#endif

	/* read the same record that we read last time and fix keys */

#ifdef NOT_NEEDED_YET
int ha_rsame_with_pos(register TABLE *form, byte * buf, int inx, byte *pos)
{
  register int error=0;
  DBUG_ENTER("ha_rsame_with_pos");

  pthread_mutex_lock(&LOCK_status);
  ha_read_key_count++;
  pthread_mutex_unlock(&LOCK_status);

  if (form->db_type == DB_TYPE_ISAM)
  {
    error=ni_rsame_with_pos((N_INFO*) form->file,buf,inx,
			    (ulong) ha_get_ptr(pos,
					       form->keyfile_info.ref_length));
#ifndef DBUG_OFF
    if (error)
      DBUG_PRINT("error",("my_errno: %d",error));
#endif
  }
  else if (form->db_type == DB_TYPE_MYISAM)
  {
    error=ni_rsame_with_pos((MI_INFO*) form->file,buf,inx,
			    ha_get_ptr(pos,
				       form->keyfile_info.ref_length));
#ifndef DBUG_OFF
    if (error)
      DBUG_PRINT("error",("my_errno: %d",error));
#endif
  }
  {
    if (!(error=ha_r_rnd(form,buf,pos)) && inx != -1)
      error=ha_rsame(form,buf,inx);
  }
  if (error)
    form->status=STATUS_NOT_FOUND;
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_rsame_with_pos */
#endif

	/* read a record with rnd-access */
	/* If pos == NullS find next record */
	/* else pos must be get using ha_info() after some read */
	/* If ha_r_rnd() = -1 then record is not in database */

int ha_r_rnd(register TABLE *form, byte * buf, byte *pos)
{
  int error;
  DBUG_ENTER("ha_r_rnd");

  pthread_mutex_lock(&LOCK_status);
  ha_read_rnd_count++;
  pthread_mutex_unlock(&LOCK_status);
  if (form->db_type == DB_TYPE_MYISAM)
  {
    if ((error=mi_rrnd((MI_INFO*) form->file,buf,pos ?
		       ha_get_ptr(pos,form->keyfile_info.ref_length) :
		       HA_OFFSET_ERROR)))
    {
      if (error == 1)
	error= -1;				/* record deleted */
      else
	error= my_errno > 0 ? my_errno : HA_ERR_END_OF_FILE;
    }
  }
  else if (form->db_type == DB_TYPE_MRG_MYISAM)
  {
    if ((error=myrg_rrnd((MYRG_INFO*) form->file,buf,pos ?
			 ha_get_ptr(pos,form->keyfile_info.ref_length) :
			 HA_OFFSET_ERROR)))
    {
      if (error == 1)
	error= -1;				/* record deleted */
      else
	error= my_errno > 0 ? my_errno : HA_ERR_END_OF_FILE;
    }
  }
  else
  {
    long rnd_pos= pos ? (long) ha_get_ptr(pos,form->keyfile_info.ref_length)
      : -1L;
    DBUG_PRINT("enter",("form: %lx  buf: %lx  pos: %ld",form,buf,rnd_pos));
    if ((error=(* ((int (*)(char *,byte *,long))(*funcs[11][(uint) form->db_type-2])))((char*) form->file,buf,rnd_pos)))
    {
      if (error == 1)
	error= -1;				/* record deleted */
      else
	error= my_errno > 0 ? my_errno : HA_ERR_END_OF_FILE;
    }
  }
#ifndef DBUG_OFF
  if (error > 0 && error != HA_ERR_END_OF_FILE)
    DBUG_PRINT("error",("error: %d  my_errno: %d",error,my_errno));
#endif
  if (error)
    form->status=STATUS_NOT_FOUND;
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_r_rnd */


	/* Read first row from a table */

int ha_readfirst(register TABLE *form, byte * buf)
{
  register int error;
  DBUG_ENTER("ha_readfirst");

  pthread_mutex_lock(&LOCK_status);
  ha_read_first_count++;
  pthread_mutex_unlock(&LOCK_status);

  error=0;
  if (!(form->db_stat & HA_OPEN_KEYFILE) || form->keys == 0 ||
      ! (ha_option_flag[form->db_type] & HA_READ_ORDER))
  {
    ha_reset(form);
    while ((error= ha_r_rnd(form,buf,(byte*) NullS)) <0) ;
    DBUG_RETURN(error);
  }
  else
    error=ha_rfirst(form,buf,0);
  if (error)
    form->status=STATUS_NOT_FOUND;
  else
    form->status=0;
  DBUG_RETURN (error);
} /* ha_readfirst */


ha_rows ha_records_in_range(TABLE *table,int inx,
			    const byte *start_key,uint start_key_len,
			    enum ha_rkey_function start_search_flag,
			    const byte *end_key,uint end_key_len,
			    enum ha_rkey_function end_search_flag)
{
  switch (table->db_type) {
  case DB_TYPE_ISAM:
    return (ha_rows) ni_records_in_range((N_INFO*) table->file,
					 inx,
					 start_key,start_key_len,
					 start_search_flag,
					 end_key,end_key_len,
					 end_search_flag);
  case DB_TYPE_MYISAM:
    return (ha_rows) mi_records_in_range((MI_INFO*) table->file,
					 inx,
					 start_key,start_key_len,
					 start_search_flag,
					 end_key,end_key_len,
					 end_search_flag);
  default:
    return 0;					// Impossible
  }
}



void ha_store_ptr(byte *buff, uint pack_length, my_off_t pos)
{
  switch (pack_length) {
#if SIZEOF_OFF_T > 4
  case 8: mi_int8store(buff,pos); break;
  case 7: mi_int7store(buff,pos); break;
  case 6: mi_int6store(buff,pos); break;
  case 5: mi_int5store(buff,pos); break;
#endif
  case 4: mi_int4store(buff,pos); break;
  case 3: mi_int3store(buff,pos); break;
  case 2: mi_int2store(buff,(uint) pos); break;
  case 1: buff[0]= (uchar) pos; break;
  }
  return;
}

my_off_t ha_get_ptr(byte *ptr, uint pack_length)
{
  my_off_t pos;
 switch (pack_length) {
#if SIZEOF_OFF_T > 4
  case 8:
    pos= (my_off_t) mi_uint8korr(ptr);
    break;
  case 7:
    pos= (my_off_t) mi_uint7korr(ptr);
    break;
  case 6:
    pos= (my_off_t) mi_uint6korr(ptr);
    break;
  case 5:
    pos= (my_off_t) mi_uint5korr(ptr);
    break;
#endif
  case 4:
    pos= (my_off_t) mi_uint4korr(ptr);
    break;
  case 3:
    pos= (my_off_t) mi_uint3korr(ptr);
    break;
  case 2:
    pos= (my_off_t) mi_uint2korr(ptr);
    break;
  case 1:
    pos= (my_off_t) mi_uint2korr(ptr);
    break;
 }
 return pos;
}


	/* Get KEYFILE_INFO from database */
	/* if flag == 1 we only get lastpos */

void ha_position(register TABLE *form)
{
  my_off_t position;
  DBUG_ENTER("ha_position");

  switch (form->db_type) {
  case DB_TYPE_MYISAM:
    {
      position=mi_position((MI_INFO*) form->file);
    }
    break;
  case DB_TYPE_ISAM:
    {
      position=ni_position((N_INFO*) form->file);
      if (position == (my_off_t) ~ (ulong) 0)
	position=HA_OFFSET_ERROR;
    }
    break;
#ifndef NO_HEAP
  case DB_TYPE_HEAP:
    {
      position=heap_position((HP_INFO*) form->file);
      if (position == (my_off_t) ~ (ulong) 0)
	position=HA_OFFSET_ERROR;
      mi_int4store(form->keyfile_info.ref, position);
      DBUG_VOID_RETURN;
    }
    break;
#endif
#ifndef NO_MERGE
  case DB_TYPE_MRG_ISAM:
    position=mrg_position((MRG_INFO*) form->file);
    break;
  case DB_TYPE_MRG_MYISAM:
    position= (my_off_t) myrg_position((MYRG_INFO*) form->file);
    break;
#endif
#ifndef NO_HASH
  case DB_TYPE_HASH:
    {
      HASHINFO info;
      VOID(h_info((H_INFO*) form->file,&info,0));
      position=info.h_recpos;
      if (position == (my_off_t) ~ (ulong) 0)
	position=HA_OFFSET_ERROR;
    }
    break;
#endif
  default:
    break;						/* impossible  */
  }
  ha_store_ptr(form->keyfile_info.ref, form->keyfile_info.ref_length,
	       position);
  DBUG_PRINT("exit",("position: %ld",(ulong) position));
  DBUG_VOID_RETURN;
} /* ha_position */


	/* Get KEYFILE_INFO from database */

void ha_info(register TABLE *form, int flag)
{
  DBUG_ENTER("ha_info");

  switch (form->db_type) {
#ifndef NO_HASH
  case DB_TYPE_HASH:
    {
      HASHINFO info;
      VOID(h_info((H_INFO*) form->file,&info,flag));
      if (flag != 1)
      {
	form->keyfile_info.records = info.h_ant;
	form->keyfile_info.deleted = info.h_delete;
	form->keyfile_info.errkey  = info.h_errkey;
	form->keyfile_info.mean_rec_length=form->reclength;
	form->keyfile_info.data_file_length=form->reclength*
	  (info.h_ant+info.h_delete);
	form->keyfile_info.ref_length=4;
      }
    }
    break;
#endif
#ifndef NO_HEAP
  case DB_TYPE_HEAP:
    {
      HEAPINFO info;
      VOID(heap_info((HP_INFO *) form->file,&info,flag));
      if (flag != 1)
      {
	form->keyfile_info.records = info.records;
	form->keyfile_info.deleted = info.deleted;
	form->keyfile_info.errkey  = info.errkey;
	form->keyfile_info.mean_rec_length=info.reclength;
	form->keyfile_info.data_file_length=info.data_length;
	form->keyfile_info.index_file_length=info.index_length;
	form->keyfile_info.max_data_file_length=info.max_records*
	  info.reclength;
	form->keyfile_info.delete_length=info.deleted*
	  info.reclength;
	form->keyfile_info.ref_length=4;
      }
    }
    break;
#endif
  case DB_TYPE_ISAM:
    {
      N_ISAMINFO info;
      VOID(ni_info((N_INFO*) form->file,&info,flag));
      if (flag != 1)
      {
	form->keyfile_info.records = info.records;
	form->keyfile_info.deleted = info.deleted;
	form->keyfile_info.data_file_length=info.data_file_length;
	form->keyfile_info.index_file_length=info.index_file_length;
	form->keyfile_info.max_data_file_length=info.max_data_file_length;
	form->keyfile_info.max_index_file_length=info.max_index_file_length;
	form->keyfile_info.delete_length = info.delete_length;
	form->keyfile_info.errkey  = info.errkey;
	form->keyfile_info.create_time = info.create_time;
	form->keyfile_info.check_time  = info.isamchk_time;
	form->keyfile_info.update_time = info.update_time;
	form->keyfile_info.mean_rec_length=info.mean_reclength;
	form->keyfile_info.sortkey = info.sortkey;
	form->keyfile_info.filenr  = info.filenr;
	form->keyfile_info.block_size=nisam_block_size;
	form->keyfile_info.ref_length=4;
	form->db_create_options    = info.options;
	/* Set keys in use */
	if (form->db_stat & HA_OPEN_KEYFILE)
	  form->keys		   = min(form->keys,info.keys);
	form->keys_in_use= (((key_map) 1) << form->keys)-1;
	ha_store_ptr(form->keyfile_info.dupp_ref,
		     form->keyfile_info.ref_length,
		     info.dupp_key_pos);
	if (!form->tmp_table)
	{
	  ulong *rec_per_key=info.rec_per_key;
	  for (uint i=0 ; i < form->keys ; i++)
	  {
	    form->key_info[i].rec_per_key[form->key_info[i].key_parts-1]=
	      *(rec_per_key++);
	  }
	}
      }
    }
    break;
#ifndef NO_MERGE
  case DB_TYPE_MRG_ISAM:
    {
      MERGE_INFO info;
      VOID(mrg_info((MRG_INFO*) form->file,&info,flag));
      if (flag != 1)
      {
	form->keyfile_info.records = (ha_rows) info.records;
	form->keyfile_info.deleted = (ha_rows) info.deleted;
	form->keyfile_info.errkey  = info.errkey;
	form->keyfile_info.data_file_length=info.data_file_length;
	form->db_create_options    = info.options;
	form->keyfile_info.mean_rec_length=info.reclength;
	form->keyfile_info.ref_length=4;
      }
    }
    break;
#endif
#ifndef NO_MERGE
  case DB_TYPE_MRG_MYISAM:
    {
      MYMERGE_INFO info;
      VOID(myrg_status((MYRG_INFO*) form->file,&info,flag));
      if (flag != 1)
      {
	form->keyfile_info.records = (ha_rows) info.records;
	form->keyfile_info.deleted = (ha_rows) info.deleted;
	form->keyfile_info.errkey  = info.errkey;
	form->keyfile_info.data_file_length=info.data_file_length;
	form->db_create_options    = info.options;
	form->keyfile_info.mean_rec_length=info.reclength;
	form->keyfile_info.ref_length=6;
      }
    }
#endif
    break;
  case DB_TYPE_MYISAM:
    {
      MI_ISAMINFO info;
      VOID(mi_status((MI_INFO*) form->file,&info,flag));
      if (flag != 1)
      {
	form->keyfile_info.records = info.records;
	form->keyfile_info.deleted = info.deleted;
	form->keyfile_info.data_file_length=info.data_file_length;
	form->keyfile_info.index_file_length=info.index_file_length;
	form->keyfile_info.max_data_file_length=info.max_data_file_length;
	form->keyfile_info.max_index_file_length=info.max_index_file_length;
	form->keyfile_info.delete_length = info.delete_length;
	form->keyfile_info.errkey  = info.errkey;
	form->keyfile_info.create_time = info.create_time;
	form->keyfile_info.check_time  = info.check_time;
	form->keyfile_info.update_time = info.update_time;
	form->keyfile_info.mean_rec_length=info.mean_reclength;
	form->keyfile_info.sortkey = info.sortkey;
	form->keyfile_info.filenr  = info.filenr;
	form->keyfile_info.block_size=nisam_block_size;
	form->keyfile_info.ref_length=info.reflength;
	form->keyfile_info.auto_increment_value= info.auto_increment;
	form->db_create_options    = info.options;
	form->keys_in_use	  &= info.key_map;
	if (form->key_parts)
	  memcpy((char*) form->key_info[0].rec_per_key, 
		 (char*) info.rec_per_key,
		 sizeof(ulong)*form->key_parts);
	ha_store_ptr(form->keyfile_info.dupp_ref,
		     form->keyfile_info.ref_length,
		     info.dupp_key_pos);
      }
    }
    break;
  default:
    break;						/* impossible  */
  }
  DBUG_PRINT("exit",("records: %ld  deleted: %ld",
		     (ulong) form->keyfile_info.records,
		     (ulong) form->keyfile_info.deleted));
  DBUG_VOID_RETURN;
} /* ha_info */


	/* Some extra flags we can use on database */

int ha_extra(TABLE *form, enum ha_extra_function operation)
{
  DBUG_ENTER("ha_extra");
  DBUG_PRINT("enter",("operation: %d",operation));

  if ((specialflag & SPECIAL_SAFE_MODE || test_flags & TEST_NO_EXTRA) &&
      (operation == HA_EXTRA_WRITE_CACHE ||
       operation == HA_EXTRA_KEYREAD))
    DBUG_RETURN(0);
  DBUG_RETURN((* ((int (*)(char *,int))(*funcs[13][(uint) form->db_type-2])))
	      ((char*) form->file,(int) operation));
} /* ha_extra */


	/* Reset database to state after open */

void ha_reset(TABLE *form)
{
  DBUG_ENTER("ha_reset");
  VOID((* ((int (*)(char *,int))(*funcs[13][(uint) form->db_type-2])))((char*) form->file,(int) HA_EXTRA_RESET));
  DBUG_VOID_RETURN;
} /* ha_reset */


	/* Lock a database for reading or writing */

int ha_lock(TABLE *form, int lock_type)
{
  int error;
  DBUG_ENTER("ha_lock");
  if (!(error=(* ((int (*)(char *,int))(*funcs[12][(uint) form->db_type-2])))((char*) form->file,lock_type)))
    form->locked=lock_type;
  DBUG_RETURN(error);
} /* ha_lock */


	/* close, flush or restart databases */
	/* Ignore this for other databases than ours */

int ha_panic(enum ha_panic_function flag)
{
  int error=0;
#ifndef NO_MERGE
  error|=mrg_panic(flag);
#endif
#ifndef NO_HASH
  error|=h_panic(flag);			/* fix hash */
#endif
#ifndef NO_HEAP
  error|=heap_panic(flag);
#endif
  error|=ni_panic(flag);
  return error;
} /* ha_panic */


	/* Use key cacheing on all databases */

void ha_key_cache(void)
{
  if (keybuff_size)
    VOID(init_key_cache(keybuff_size,(uint) (10*4*(IO_SIZE+MALLOC_OVERHEAD))));
} /* ha_key_cache */


	/* Print error that we got from handler function */

void ha_error(TABLE *form, int error, myf errflag)
{
  int textno=ER_GET_ERRNO;
  switch (error) {
  case EAGAIN:
    textno=ER_FILE_USED;
    break;
  case ENOENT:
    textno=ER_FILE_NOT_FOUND;
    break;
  case HA_ERR_KEY_NOT_FOUND:
  case HA_ERR_NO_ACTIVE_RECORD:
  case HA_ERR_END_OF_FILE:
    textno=ER_KEY_NOT_FOUND;
    break;
  case HA_ERR_FOUND_DUPP_KEY:
  {
    int key_nr=ha_keyerror(form,error);
    if (key_nr >= 0)
    {
      /* Write the dupplicated key in the error message */
      char key[MAX_KEY_LENGTH];
      String str(key,sizeof(key));
      key_unpack(&str,form,(uint) key_nr);
      uint max_length=MYSQL_ERRMSG_SIZE-strlen(ER(ER_DUP_ENTRY));
      if (str.length() >= max_length)
      {
	str.length(max_length-4);
	str.append("...");
      }
      my_error(ER_DUP_ENTRY,MYF(0),str.c_ptr(),key_nr+1);
      return;
    }
    textno=ER_DUP_KEY;
    break;
  }
  case HA_ERR_FOUND_DUPP_UNIQUE:
    textno=ER_DUP_UNIQUE;
    break;
  case HA_ERR_RECORD_CHANGED:
    textno=ER_CHECKREAD;
    break;
  case HA_ERR_CRASHED:
    textno=ER_NOT_KEYFILE;
    break;
  case HA_ERR_OUT_OF_MEM:
    my_error(ER_OUT_OF_RESOURCES,errflag);
    return;
  case HA_ERR_WRONG_COMMAND:
    textno=ER_ILLEGAL_HA;
    break;
  case HA_ERR_OLD_FILE:
    textno=ER_OLD_KEYFILE;
    break;
  case HA_ERR_UNSUPPORTED:
    textno=ER_UNSUPPORTED_EXTENSION;
    break;
  case HA_ERR_RECORD_FILE_FULL:
    textno=ER_RECORD_FILE_FULL;
    break;
  default:
    {
      my_error(ER_GET_ERRNO,errflag,error);
      return;
    }
  }
  my_error(textno,errflag,form->table_name,error);
} /* ha_error() */


	/* Return key if error because of dupplicated keys */

int ha_keyerror(TABLE *form, int error)
{
  DBUG_ENTER("ha_keyerror");
  form->keyfile_info.errkey  = -1;
  if (error == HA_ERR_FOUND_DUPP_KEY || error == HA_ERR_FOUND_DUPP_UNIQUE)
    ha_info(form,2);
  DBUG_RETURN(form->keyfile_info.errkey);
} /* ha_keyerror */


	/* Use other databasehandler if databasehandler is not incompiled */

enum db_type ha_checktype(enum db_type database_type)
{
  switch (database_type) {
#ifndef NO_HASH
  case DB_TYPE_HASH:
#endif
#ifndef NO_HEAP
  case DB_TYPE_HEAP:
#endif
#ifndef NO_MERGE
  case DB_TYPE_MRG_ISAM:
  case DB_TYPE_MRG_MYISAM:
#endif
  case DB_TYPE_ISAM:
  case DB_TYPE_MYISAM:
    return (database_type);			/* Database exists on system */
  default:
    break;
  }
  return(DB_TYPE_ISAM);				/* Use this as default */
} /* ha_checktype */



	/* rename database (indexfile and datafile) */

int ha_frename(enum db_type base, const char * from, const char * to)
{
  DBUG_ENTER("ha_frename");

  if ((bas_ext[(uint) base][0][0] &&
       rename_file_ext(from,to,bas_ext[(uint) base][0])) ||
      (bas_ext[(uint) base][1][0] &&
       rename_file_ext(from,to,bas_ext[(uint) base][1])))
    DBUG_RETURN(1);
  DBUG_RETURN(0);
}

	/* dummys for not existing functions */

static int ha_err(void)
{
  my_errno=HA_ERR_WRONG_COMMAND;
  return -1;
}

static int ha_ok(void)
{
  return 0;
}

static int delete_file(my_string name,my_string ext,int extflag)
{
  char buff[FN_REFLEN];
  VOID(fn_format(buff,name,"",ext,extflag | 4));
  return(my_delete(buff,MYF(MY_WME)));
}


int ha_fdelete(enum db_type base,my_string name)
{
  DBUG_ENTER("ha_fdelete");

  if ((bas_ext[base][0][0] && delete_file(name,bas_ext[base][0],2)) ||
      (bas_ext[base][1][0] && delete_file(name,bas_ext[base][1],2)))
   DBUG_RETURN(1);

  switch (base) {
#ifndef NO_HEAP
  case DB_TYPE_HEAP:
    {
      char buff[FN_REFLEN];
      DBUG_RETURN (heap_delete_all(name));
    }
#endif
  default:
    break;
  }
  DBUG_RETURN(0);
} /* ha_fdelete */


	/* Set a timestamp in record */

static void NEAR_F set_form_timestamp(TABLE *table,byte *record)
{
  long skr= (long) current_thd->query_start();
#ifdef WORDS_BIGENDIAN
  if (table->db_low_byte_first)
  {
    int4store(record,skr);
  }
  else
#endif
  longstore(record,skr);
  return;
}


	/* Updates field with field_type NEXT_NUMBER according to following:
	   if field = 0 change field to the next free key in database.
	   */

void update_next_number(TABLE *form)
{
  longlong nr;
  int error;
  THD *thd;
  DBUG_ENTER("update_next_number");

  if (form->next_number_field->val_int() != 0)
    DBUG_VOID_RETURN;
  thd=current_thd;
  if ((nr=thd->next_insert_id))
    thd->next_insert_id=0;			// Clear after use
  else
  {
    if (form->db_type == DB_TYPE_MYISAM)
    {
      ha_info(form,2);
      nr=form->keyfile_info.auto_increment_value;
    }
    else
    {
      VOID(ha_extra(form,HA_EXTRA_KEYREAD));
      error=ha_rlast(form,form->record[1],(int) form->next_number_index);
      if (error)
	nr=1;
      else
	nr=(longlong) form->next_number_field->
	  val_int_offset(form->rec_buff_length)+1;
      VOID(ha_extra(form,HA_EXTRA_NO_KEYREAD));
    }
  }
  thd->insert_id((ulonglong) nr);
  form->next_number_field->store(nr);
  DBUG_VOID_RETURN;
}
